"""Generic reloaders."""

from abc import ABC, abstractmethod

import log

from ..utils.log import Logger


class Reloader(ABC):  # pylint: disable-msg=too-few-public-methods
    """Reloader interface."""

    @abstractmethod
    async def onConfigUpdate(self, newConfig: "BasePythonServiceSettingsClass") -> None:
        """Main method for handling configuration update."""


class LoggingReloader(Reloader):  # pylint: disable-msg=too-few-public-methods
    """Reloader for Logger configuration."""

    def __init__(self, config: "BasePythonServiceSettingsClass"):
        """
        Args:
            config: initial configuration
        """
        self.config = config
        self.logger = Logger("logging reloader")

    async def onConfigUpdate(self, newConfig: "BasePythonServiceSettingsClass"):
        """
        Check whether logger configuration changed and perform logging
        reinitialization with new configuration.

        Args:
            newConfig: new configuration
        """

        if self.config.logger == newConfig.logger:
            return

        self.logger.info("Config has changed, reloading")

        # Log suffix never change
        logSuffix = Logger._settings.logNameSuffix  # pylint:disable=W0212

        log.initiate(
            appName=newConfig.serviceName,
            logLevel=newConfig.logger.logLevel,
            logTime=newConfig.logger.logTime,
            folderForLog=newConfig.logger.folderWithLogs,
            maxSize=newConfig.logger.maxLogFileSize,
            logNameSuffix=logSuffix,
            logToStdOut=newConfig.logger.logToStdOut,
            logToFile=newConfig.logger.logToFile,
            multilineStackTrace=newConfig.logger.multilineStackTrace,
            formatter=newConfig.logger.format,
        )
        self.config = newConfig


class SanicReloader(Reloader):
    """Reloader for sanic applications."""

    def __init__(self, process: "Process", config: "BasePythonServiceSettingsClass"):
        """
        Args:
            process: Process instance
            config: initial configuration
        """
        self.config = config
        self.process = process
        self.logger = Logger("sanic reloader")

    async def onConfigUpdate(self, newConfig: "BasePythonServiceSettingsClass"):
        """
        Restart application on any change.

        Args:
            newConfig: new configuration
        """
        if self.config != newConfig:
            self.logger.info("Config has changed, restarting")
            self.config = newConfig
            self.restart()

    def restart(self):
        """Perform process restart."""
        self.process.restart()
